package ClienteServidor;

import java.net.*;
import java.io.*;
import java.applet.*;
import java.awt.*;
import java.awt.event.*;
/**
* Esta clase es la que se encarga de conectarse como un cliente.
* Brinda todos lo metodos para mandar y recibir mensajes a un servidor, asi
* como dispara todos los mensajes que le pueden llegar a un cliente
* para que una interfas grafica se los presente al usario y habilite
* y deshabilite los controles de manejo.
* @author 	Alonso Ramrez
* @version 	1.0
* @since       JDK1.0
*/		
public class ConectorCliente implements Runnable { 

	/**
	* La direccion por dafault de conexion de un socket
	*/
	public static final int default_port=6789;
	/**
	* El socket con el cual se va a conectar
	*/
	private Socket s;
	/**
	* El flujo de entrada para leer del socket
	*/
	private BufferedReader sin;
	/**
	* El flujo de salida para escribir al socket
	*/
	private PrintStream sout;
	/**
	* El login del usuario
	*/
	private String login;
	
	/**
	* Variable para detener la iteracion del Thread
	*/
	private boolean pleaseStop=false;
	
	/**
	* El objeto de interfas grafica al cual le vamos a mandar los mensajes
	* del cliente, para que se los muestre al usuario
	*/
	RecibidorMensajesCliente interfasGrafica;
	
	/**
	* Nos inndica si esta o no conectado
	*/
	public boolean conectado = false;
	
	/**
	* El hilo para que se pueda siempre estar esperando respuesta del servidor
	* sin bloquear el sistema
	*/	
	private Thread th;
	
	/**
	* Para crear un conector cliente se le deben de mandar:
	* @param login      el login
	* @param password   el password
	* @param host       (el id del servidor)
	* @param port       el puerto del socket
	* @param interfasG  La interfas grafica de usuario a donde entragara mensajes de cliente
	*/
	public ConectorCliente(String login,String password,String host,int port, RecibidorMensajesCliente interfasG) {
		interfasGrafica  = interfasG;
		if(port==0)
			construct(login,password,host,default_port);
		else
			construct(login,password,host,port);
	}
	
	/**
	* Ejecuta el codigo necesario para pedirle la hora al servidor
	*/
	public void pideHora(){
		send(Protocolo.TK_MANDA_HORA + Protocolo.chk("") + ":");
	}	
	/**
	* Ejecuta el codigo necesario para pedirle el clima al servidor
	*/
	public void pideClima(){
		send(Protocolo.TK_MANDA_CLIMA + Protocolo.chk("") + ":");
	}	
	/**
	* Ejecuta el codigo necesario para pedirle la lista de los clientes conectados
	* al servidor
	*/	
	public void pideListaClientes(){
		send(Protocolo.TK_MANDA_CLIENTES + Protocolo.chk("") + ":");
	}	
	
	/**
	* Ejecuta el codigo necesario para pedirle la lista de archivos que tiene
	* disponible el servidor para mandar
	*/
	public void pideListaArchivos(){
		send(Protocolo.TK_MANDA_LISTA_ARCH + Protocolo.chk("") + ":");
	}
	/**
	* Ejecuta el codigo necesario para pedirle un archivo al servidor
	*/	
	public void pideArchivo(String nombre){
		send(Protocolo.TK_RECIBE_ARCH + Protocolo.chk("") + ":"+nombre);
	}	
	
	
	/**
	* Ejecuta el codigo necesario para hacer un logout del servidor
	*/	
	public void hacerLogOut(){
		send(Protocolo.TK_LOGOUT + Protocolo.chk("") + ":");
		pleaseStop = true;
		th.stop();
	}

	
	/**
	* Ejecuta el codigo necesario para terminar una conexion
	*/	
	public void desconectar(){
		pleaseStop = true;
		th.stop();
	}
	
	

	/**
	* Auxiliar de construccion, asigna los streams al socket e inicializa la cominicacion
	* ya propiamente de protocolo con el servidor
	* @param login      el login
	* @param password   el password
	* @param host       (el id del servidor)
	* @param port       el puerto del socket
	*/
	private void construct(String login,String password,String host,int port) {
		this.login=login;
		try {
			
			s=new Socket(host,port);
			sin = new BufferedReader(new InputStreamReader(s.getInputStream()));
			sout= new PrintStream(s.getOutputStream());
			
			
			/*mandar Login y password*/
			
			sout.println(Protocolo.TK_LOGIN+Protocolo.chk("")+":"+login+","+password);

			//y esperar a que conteste el servidor
			th=new Thread(this); 
			th.start();
			
			
		} catch (IOException e) {
			interfasGrafica.noPudeConectar("No pude conectar: "+ e.toString());
			System.err.println(e);
		}
	}
	
	/**
	* escribe una cadena al stream de salida del socket
	*/
	void send(String msg) {
		if (msg==null||msg.length()==0) {
		} else {
			sout.println(msg);
		}
	}
	
	/*
	* Manda un mensaje a la salida estandar del sistema
	*/
	void println(String msg) {
		System.out.println(msg);
	}

	/**
	* Es la funcion del hilo que esta constantemente escuchando por si llega algun mensaje
	* asincrono del servidor, y le da matenimiento a este disparando los mensajes a la interfas de usuario
	*/
	public void run() {
		while (!pleaseStop) {
			try {
				// leer un mensaje del servidor
				String line=sin.readLine();
				
				if (line==null) { 
					System.out.println("Connection closed by server."); // algo fallo
					break;
				}
				if(conectado==false){ // si es la primera vez que nos "hablamos"?

					if(line.startsWith(Protocolo.TK_LOGIN_OK)){ // el servidor contesta que mi login fue correcto
						conectado=true;
						interfasGrafica.conectado(login+" conectado a: "+s.getInetAddress().getHostName()+" : " + s.getPort());
					}
					else if(line.startsWith(Protocolo.TK_LOGIN_MAL)){ // el servidor contesta que mi login no existe
						interfasGrafica.noPudeConectar("Desconectado: Login incorrecto");
						break;
					}
					else if(line.startsWith(Protocolo.TK_PASSWORD_MAL)){// el servidor contesta que mi password es incorrecto
						interfasGrafica.noPudeConectar("Desconectado: Password incorrecto");
						break;
					}
				}
				else{ // ya hay conexion con servidor, esta es informacion que manda el servidor
					// en cada caso hay que mandar la informacion a la interfas grafica de usuario
					if(line.startsWith(Protocolo.TK_MANDA_HORA)){ // el servidor manda la hora
						interfasGrafica.llegoInformacion(line.substring(5));
					}
					else if(line.startsWith(Protocolo.TK_MANDA_CLIMA)){// el servidor manda el clima
						interfasGrafica.llegoInformacion(line.substring(5));						
					}
					else if(line.startsWith(Protocolo.TK_MANDA_CLIENTES)){// el servidor manda la lista de clientes
						interfasGrafica.llegoListaClientes(line.substring(5));
					}
					else if(line.startsWith(Protocolo.TK_MANDA_LISTA_ARCH)){// el servidor manda la lista de archivos
						interfasGrafica.llegoListaArchivos(line.substring(5));
					}
					else if(line.startsWith(Protocolo.TK_RECIBE_ARCH)){// el servidor manda un archivo
						interfasGrafica.llegoArchivo(line.substring(5));						
					}					
					else if(line.startsWith(Protocolo.TK_SERVIDOR_ABAJO)){// el servidor se apago o se desconecto
						interfasGrafica.servidorAbajo();						
					}
					else if(line.startsWith(Protocolo.TK_MENSAJE_A_TODOS)){// el servidor mando un mensaje a todos
						interfasGrafica.mensajeATodos(line.substring(5));
					}
				}
			} catch (IOException e) {
			} finally { 
			}
		}
		try {
			if (s!=null) {
				s.close(); // no hay nada , salimos
			}
		} catch (IOException e2) {
		}
	}
}

