package ClienteServidor;

import java.net.*;
import java.io.*;
import java.util.*;
/**
* Esta clase es la que se encarga de conectarse como servidor.
* Brinda todos lo metodos para captar nuevos clientes, asi
* como dispara todos los mensajes que le pueden llegar a un servidor
* para que una interfas grafica se los presente al usario y habilite
* y deshabilite los controles de manejo.
* @author 	Alonso Ramrez
* @version 	1.0
* @since       JDK1.0
*/		
public class ConectorServidor extends Thread { 

	/** 
	* puerto default usado en la conexin al definir el socket de servidor 
	*/
	public static final int default_port=6789;
	/**
	* El no. de puerto usado para conectarse
	*/
	protected int port;
	
	/** 
	* Socket de servidor usado en las comunicaciones 
	*/
	protected ServerSocket listen_socket;
	
	/** 
	* vector de los clientes conectados al servidor 
	*/
	Vector clients;
	/**
	* flag que le indica al thread que debe de terminar
	*/
	public boolean pleaseStop = false;
	
	/**
	* Nos indica si actualmente hay conexion
	*/
	public boolean conectado = false;
	
	/**
	* La interfas grafica de usuario, a la cual le mandamos los mensajes de un servidor
	*/
	RecibidorMensajesServidor interfasGrafica;

	/**
	* Sale de la aplicacion si hay problemas con el socket
	*/
	public static void fail(Exception e,String msg) { 
		System.err.println(msg+":"+e);
		System.exit(1);
	}
	
	
	/**
	* Se desconecta avisando a todos los clientes que ya no esta disponible
	*/
	public void desconectar(){
		//barrer todo el vector y mandar un mensaje de desconexion
		broadcast(Protocolo.TK_SERVIDOR_ABAJO+Protocolo.chk("")+":",null);
		pleaseStop = true;
		stop();
		try{
			listen_socket.close();
		}
		catch(IOException ioe){
			
		}
		clients.removeAllElements(); 

	
	}	
	/** construye un servidor que escucha en el puerto port 
	*@port         nmero del puerto
	*@pInterGraf   la interfas grafica a la cual se le mandan mensajes
	*/
	public ConectorServidor(int port,RecibidorMensajesServidor pInterGraf) {
		
		if (port==0) {
			// se acepta default
			port=default_port;
		}
		
		// guarda interfas grafica
		interfasGrafica= pInterGraf;
		
		this.port=port;
		try { 
			// crea el socket servidor
			listen_socket=new ServerSocket(port); 
		} catch (IOException e) {
			// no se pudo crear el socket
			interfasGrafica.noPudeConectar("Exception Creating Server Socket");
			fail(e,"Exception Creating Server Socket");
			return;
		}
		// ya hay conexion ok
		System.out.println("Server: listening on port "+port);
		interfasGrafica.conectado("Servidor escuchando en puerto "+ port);
		
		// indicar que tenemos conexion
		conectado = true;
		
		// crear el vector listo para ir agregando clientes
		clients=new Vector();

		// empezar el hilo para esperar clientes		
		this.start();
	}
	
	/**
	* La funcion del thread que esta siempre esperando conexion de nuevos clientes
	*/
	public void run() {
		try {
			while (!pleaseStop) {
				// estar esperando el arribo de un nuevo cliente
				Socket client_socket=listen_socket.accept();
				// crear una nueva conexion para este cliente
				myConnectionServer c=new myConnectionServer(this,client_socket);
				
				// agregarlo a la,lista de clientes
				clients.addElement(c);
				
				// avisar a la interfas grafica que llego un nuevo cliente
				interfasGrafica.llegoClienteNuevo("Llego un cliente nuevo");
			}
		}
		catch (IOException e) {
			fail(e,"Exception while listening for connections");
		}
	}
	
	/** 
	*dar de baja la conexin mc (a un cliente) 
	*@param mc   la conexion a dar de baja
	*/
	void remove(myConnectionServer mc) {
		clients.removeElement(mc);
	}
	
	/** 
	*enviar a todos los clientes un mensaje 
	*@param msg el mensaje
	*@param mc0 la conexin de donde viene
	*/
	void broadcast(String msg,myConnectionServer mc0) {
		for (int i=0; i<clients.size();i++) {
			myConnectionServer mc=(myConnectionServer) clients.elementAt(i);
			mc.println(msg);
			System.out.println("mande "+msg+","+mc.id);
		}
	}
	
	/**
	* Regresa en un String la lista de clientes separados por comas
	*/
	String listaClientes(){
		String regreso = new String("");
		for (int i=0; i<clients.size();i++) {
			myConnectionServer mc=(myConnectionServer) clients.elementAt(i);
			regreso = regreso.concat(mc.id+",");
		}
		return regreso;
		
	}
	
	/**
	* Envia un mensaje a todos los clientes conectados (con Protocolo)
	*@param mensage          el mensaje
	*/
	public void mandaMensajeATodos(String mensaje){
		broadcast(Protocolo.TK_MENSAJE_A_TODOS+Protocolo.chk("")+":"+mensaje,null);
	}
	
		
}

		
/**
* Es la encargada de manejar la coneccion con un solo cliente, darle amtenimiento a todos
* los mensajes que este puede mandar, asi como contestarle
* @author 	Alonso Ramrez
* @version 	1.0
* @since       JDK1.0
*/		
class myConnectionServer extends Thread {
	
	/**
	* El socket de este cliente
	*/
	protected Socket client;
	/**
	* El flujo de entrada para este socket
	*/
	protected BufferedReader in;
	/**
	* El flujo de salida para este socket
	*/
	protected PrintStream out;
	
	/**
	* El servidor
	*/
	private ConectorServidor ms;
	
	/**
	* Indica si esta coneccion ya fue validada con login y password, a apenas se comiensa el
	* protocolo
	*/
	protected boolean validada=false;
	
	/**
	* El id de este cliente conectado
	*/
	protected String id;
	
	
	/**
	* Para crear una conexion se le manda el apuntador al servidor y el apuntador
	* al cliente que se quiere conectar
	*/
	public myConnectionServer(ConectorServidor ms,Socket client_socket) {
		// guardar datos
		this.ms=ms;
		client= client_socket;
		
		try {
			// asignar los flujos de I/O
			in=new BufferedReader(new InputStreamReader(client.getInputStream()));
			out=new PrintStream(client.getOutputStream());
		} catch (IOException e) {
			try {
				client.close(); 
			} catch (IOException e2) {
			}
			// error al asignar streams
			System.err.println("Exception while getting socket streams: "+e);
			return;
		}
		// empezar a escuchar a este cliente para contestarle
		this.start();
	}
	
	/**
	* Escribe un mensaje a la salida de la consola
	*/
	void println(String msg) {
		out.println(msg);
	}
		
	/**
	* Le da matenimiento a todos los requerimientos de un cliente
	*/
	public void run() {
		String line="";
		StringBuffer revline;
		int len;
		
		try {
			for (;;) { 
				//leer una peticion
				line = in.readLine();
				if (line==null||line.length()==0 || line.startsWith(Protocolo.TK_LOGOUT)) {
					// hizo log out el cliente o bien ya se murio
					break;
				}
				else if(!validada){
					// aqui entra cuando se esta dando de alta el cliente
					if(!line.startsWith(Protocolo.TK_LOGIN)){
						ms.interfasGrafica.noSeIdentificoClienteNuevo("anonimo");
						break; // no es el protocolo cancelar
					}
					//buscarlo en la base de datos
					int r = estaEnBaseDeDatos(line);
					if(r ==0){
						// existe en la base de datos y el password es correcto
						validada=true; // ya esta validadada
						out.println(Protocolo.TK_LOGIN_OK+Protocolo.chk("")+":");
						ms.interfasGrafica.seDioDeAltaunClienteNuevo(id);
					}
					else if (r ==-1 || r == -2){
						// no esta en la base de datos o la trama de login era incorrecta
						out.println(Protocolo.TK_LOGIN_MAL+Protocolo.chk("")+":");
						ms.interfasGrafica.noSeIdentificoClienteNuevo(line+" (login incorrecto)");
						id=line;
						break; //cancelar conexion
					}
					else if (r ==-3){
						// el password no coincide
						out.println(Protocolo.TK_PASSWORD_MAL+Protocolo.chk("")+":");
						ms.interfasGrafica.noSeIdentificoClienteNuevo(line+" (password incorrecto)");
						// guardar lo que llego para mostrarlo en pantalla
						id=line;
						break; //cancelar conexion
					}
				}
				else{ // todos los demas mensajes, peticiones de cliente
				
					if(line.startsWith(Protocolo.TK_MANDA_HORA)){
						//el cliente quiere saber la hora
						Date d = new Date();
						out.println(Protocolo.TK_MANDA_HORA+Protocolo.chk("")+":En el servidor son las "+d.getHours()+" hrs. con "+d.getMinutes()+" mins. y "+d.getSeconds()+" segs.");
						ms.interfasGrafica.accionCliente(id,"Pidio la hora");

					}
					else if(line.startsWith(Protocolo.TK_MANDA_CLIMA)){ 
						// el cliente quiere saber el clima
						//contestarle
						out.println(Protocolo.TK_MANDA_CLIMA+Protocolo.chk("")+":En el lugar donde esta el servidor el clima es: "+clima());							
						// avisar a la interfas grafica
						ms.interfasGrafica.accionCliente(id,"Pidio el clima");
					}
					else if(line.startsWith(Protocolo.TK_MANDA_CLIENTES)){
						// el cliente quiere saber la lista de clientes
						String lista = ms.listaClientes();
						out.println(Protocolo.TK_MANDA_CLIENTES+Protocolo.chk("")+":"+lista);							
						ms.interfasGrafica.accionCliente(id,"Pidio Lista de Clientes");

					}
					else if(line.startsWith(Protocolo.TK_MANDA_LISTA_ARCH)){
						// el cliente quiere saber la lista de archivos
						// obtener usar un listador de Archivos
						ListaDeArchivos lArch=new ListaDeArchivos();
						lArch.ext=".trf"; // los archivos trasfer file (TRF)
						lArch.obtiene();
						String lista="";
						
						if(lArch.files == null || lArch.files.length == 0)
							System.out.println("No hay archivos");
						else{
							for (int i=0;i<lArch.files.length;i++) {
								lista = lista.concat(lArch.files[i]+",");
								System.out.println(lArch.files[i]);
							}
						}
						out.println(Protocolo.TK_MANDA_LISTA_ARCH+Protocolo.chk("")+":"+lista);	
						ms.interfasGrafica.accionCliente(id,"Pidio lista de Archivos");
												
					}
					else if(line.startsWith(Protocolo.TK_RECIBE_ARCH)){
						// el cliente quiere recibir un archivo
						out.println(Protocolo.TK_RECIBE_ARCH+Protocolo.chk("")+":"+regresaArchivo(line.substring(5)));
						ms.interfasGrafica.accionCliente(id,"Pidio el archivo "+ line.substring(5) );
						
					}
				}
				
			}
		} catch (IOException e) {
			System.out.println("exeption");
			
		} finally {
			try { 
				// como ya se murio o hizo logout hay que destruirlo
				client.close(); 
				// quitarlo de la lista
				ms.remove(this);
				// avisar a la interfas grafica
				ms.interfasGrafica.seDioDeBajaCliente(id);
			} catch(IOException e2) {
			}
		}
	}
	
	/**
	* Genera un clima aleatorio (a falta de un super sensor :-) )
	*/
	private String clima(){
		//genrar un clima a leatorio
		Random a = new Random();
		if(a.nextDouble()<0.25)
			return "Soleado";
		if(a.nextDouble()<0.5)
			return "Nublado";
		if(a.nextDouble()<0.75)
			return "Lluvioso";
		return "Templado";
	}

	/**
	* Entrega el contenido completo de un archivo en un String
	* @param Archivo   el nombre del archivo
	*/
	private String regresaArchivo(String Archivo){

		try{
			// abrirlo
			RandomAccessFile raf=new RandomAccessFile(Archivo,"r");
			byte[] b=new byte[(int)raf.length()];
			// leerlo en un array de btes
			raf.readFully(b);
			
			// eliminar los enters por un codigo propio
			for(int i=0;i<raf.length();i++){
				System.out.println(">"+(int)b[i]+"<\n");
				if((int)b[i] == 13)
					b[i]=(byte)'~';
				else if((int)b[i] == 10)
					b[i]=(byte)'$';

			}
			raf.close();
			//convertir el array de bytes a cadena
			return new String(b);
		}
		catch (IOException e){
			System.out.println(e+"->" +e.getMessage());
			return "";
		}
	}
	
	/**
	* Nos indica si un login es valido, es decir si esta en la tabla
	* de usuarios. Si es asi, respalda el login para identificar al cliente.
	* @param sLogin  La cadena que tiene el login y password
	* @return   <code>0</code>  Login Ok
	*			<code>-1</code> No es trama de login valida
	*			<code>-2</code> No esta en la tabla
	*			<code>-3</code> El password no es correcto
	*/
	private int estaEnBaseDeDatos(String sLogin){
		// indica donde termina el login
		int separador=sLogin.indexOf(",");
		// donde empieza la info
		int inicio =sLogin.indexOf(":");
		// el numero de registros de la base de datos
		long totRegs=0;
		// el nombre del archivo de tabla de usuarios
		String nomArch = "Base.dat";
		// la longitud del registro de la tabla de usuario
		int lenReg=16;
		
		
		if(separador == -1 || inicio ==-1 )
			// no es una trama valida
			return -1;
			
		//Sacar Login
		String id = sLogin.substring(inicio+1   ,separador);
		while(id.length()<8) //rellenar a 8 caracteres
			id = id.concat(" ");
			
		//Sacar passw
		String pw = sLogin.substring(separador+1,sLogin.length());
		while(pw.length()<8) //rellenar a 8 caracteres
			pw = pw.concat(" ");

		
		try{
			// abrir el archivo
			RandomAccessFile raf=new RandomAccessFile(nomArch,"r");
			
			
			totRegs=raf.length()/lenReg;
			
			if(totRegs==0)
				return -2;
				
			byte lectura[]=new byte[8];
			String unID,unPW;
			
			// barrer toda la tabla de usuarios
			for(int i =0 ; i < totRegs; i++){
				raf.read(lectura,0,8);
				unID = new String(lectura);
				
				raf.read(lectura,0,8);
				unPW = new String(lectura);

				raf.read(lectura,0,2);// quitar CR LF
	
				if(unID.equals(id)) // id ok
					if(unPW.equals(pw)){ //passw ok
						raf.close();
						//guardar el id
						this.id = id;
						return 0; //todo ok;
					}
					else{
						raf.close();
						return -3;
					}
					
			}
			raf.close();

		}
		catch (IOException e){
			System.out.println(e+"->" +e.getMessage());
			return -2;
		}

		return -2; //no existe en la tabla
		
	}
		
}

